var map;

$(document).ready(function() {
	var resize = function () {
		var $map = $('#map');
		
		$map.height($(window).height() - $('div.navbar').outerHeight());
		
		if (map) {
			map.invalidateSize();
		}
	};
	
	$(window).on('resize', function () {
		resize();
	});
	
	resize();
	
	map = L.map('map').setView([-4.0, 13.0], 6);

	var baseLayer = L.tileLayer('http://{s}.tile.cloudmade.com/82e1a1bab27244f0ab6a3dd1770f7d11/999/256/{z}/{x}/{y}.png', {
	    attribution: 'Map data &copy; <a href="http://openstreetmap.org">OpenStreetMap</a> contributors, <a href="http://creativecommons.org/licenses/by-sa/2.0/">CC-BY-SA</a>, Imagery  <a href="http://cloudmade.com">CloudMade</a>'
	});
	
	baseLayer.addTo(map);
	
	var layerControl = L.control.layers({
		'Cloudmade (Dark)': baseLayer 
	}).addTo(map);
	
	var marker;
	var layer;
	
	var createLayerGroup = function (name) {
		var layerGroup = new L.LayerGroup();
		
		map.addLayer(layerGroup);
		layerControl.addOverlay(layerGroup, name);
		
		return layerGroup;
	};
	
	var addMarkers = function (layerGroupName, lat, lng, deltaLng, count, markerFunction, text) {
		
		var layerGroup = createLayerGroup(layerGroupName);
		
		var callout = new L.Callout(new L.LatLng(lat, lng), {
			direction: L.CalloutLine.DIRECTION.NW,
			lineStyle: L.CalloutLine.LINESTYLE.STRAIGHT,
			numberOfSides: 3,
			arrow: true,
			color: '#C0C0C0',
			fillColor: '#C0C0C0',
			position: new L.Point(-60, 0),
			size: new L.Point(40, 0),
			icon: new L.DivIcon({
				iconSize: new L.Point(80, 50),
				html: '<div>' + layerGroupName + '</div>',
				className: 'callout-text'
			})
		});
	
		layerGroup.addLayer(callout);
			
		for (var i = 0; i < count; ++i) {
			layerGroup.addLayer(markerFunction(new L.LatLng(lat, lng + i * deltaLng), i));
		}
	};
	
	// Create some sample basic shape markers.  NOTE:  There are built-in classes for constructing the common shapes:
	// triangles, squares, pentagons, hexagons, and octagons that can be used instead of the L.RegularPolygonMarker class.			
	addMarkers('Regular Polygons', 0.0, 0.0, 2.0, 5, function (latlng, index) {
		var colorValue = index * 20;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: 20,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			numberOfSides: index + 3
		};
		
		options.rotation = (options.numberOfSides % 2 === 0 ? 180 : 90)/options.numberOfSides;
		
		return new L.RegularPolygonMarker(latlng, options);
	});
	
	// Create versions of the basic regular polygons with holes
	addMarkers('Regular Polygons (Hollow)', -2.0, 0.0, 2.0, 5, function (latlng, index) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: Math.random() * 40 + 10,
			fillOpacity: 0.7,
			rotation: Math.random() * 360,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			numberOfSides: index + 3
		};
		
		options.innerRadius = options.radius/2;

		return new L.RegularPolygonMarker(latlng, options);
	});
	
	addMarkers('Stars', -4.0, 0.0, 2.0, 5, function (latlng, index) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: Math.random() * 40 + 10,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			rotation: 55,
			numberOfPoints: index + 5
		};
		
		options.innerRadius = options.radius/2;

		return new L.StarMarker(latlng, options);
	});
	
	addMarkers('Map Markers', -6.0, 0.0, 2.0, 5, function (latlng, index) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			radius: (Math.random() * 15) + 5
		};
		
		options.innerRadius = options.radius/ ((Math.random() * 3) + 2);
		
		return new L.MapMarker(latlng, options);
	});
	
	addMarkers('Radial Meter Markers', -8.0, 0.0, 2.0, 5, function (latlng, index) {
		var minHue = 120;
		var maxHue = 0;
		var meterMarkerOptions = {
			data: {
				'Speed': Math.random() * 200
			},
			chartOptions: {
				'Speed': {
					displayName: 'Speed',
					displayText: function (value) {
						return value.toFixed(1);
					},
					color: 'hsl(240,100%,55%)',
					fillColor: 'hsl(240,80%,55%)',
					maxValue: 200,
					minValue: 0
				}
			},
			displayOptions: {
				'Speed': {
					color: new L.HSLHueFunction(new L.Point(0,minHue), new L.Point(200,maxHue), {outputSaturation: '100%', outputLuminosity: '25%'}),
					fillColor: new L.HSLHueFunction(new L.Point(0,minHue), new L.Point(200,maxHue), {outputSaturation: '100%', outputLuminosity: '50%'})
				}
			},
			fillOpacity: 0.8,
			opacity: 1,
			weight: 0.5,
			radius: 30,
			barThickness: 15,
			maxDegrees: 360,
			rotation: 0,
			numSegments: 10
		};

		return new L.RadialMeterMarker(latlng, meterMarkerOptions);
	});
	
	addMarkers('Bar Charts', 0.0, 14.0, 2.0, 5, function (latlng) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: 20,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			width: 8
		};
		
		options.data = {
			'dataPoint1': Math.random() * 20,
			'dataPoint2': Math.random() * 20,
			'dataPoint3': Math.random() * 20,
			'dataPoint4': Math.random() * 20
		};
		
		options.chartOptions = {
			'dataPoint1': {
				fillColor: '#F2F0F7',
				minValue: 0,
				maxValue: 20,
				maxHeight: 30,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint2': {
				fillColor: '#CBC9E2',
				minValue: 0,
				maxValue: 20,
				maxHeight: 30,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint3': {
				fillColor: '#9E9AC8',
				minValue: 0,
				maxValue: 20,
				maxHeight: 30,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint4': {
				fillColor: '#6A51A3',
				minValue: 0,
				maxValue: 20,
				maxHeight: 30,
				displayText: function (value) {
					return value.toFixed(2);
				}
			}
		};
		
		return new L.BarChartMarker(latlng, options);
	});
	
	addMarkers('Pie Charts', -2.0, 14.0, 2.0, 5, function (latlng) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: 20,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			numberOfSides: 50,
			barThickness: 10
		};

		options.data = {
			'dataPoint1': Math.random() * 20,
			'dataPoint2': Math.random() * 20,
			'dataPoint3': Math.random() * 20,
			'dataPoint4': Math.random() * 20
		};
		
		options.chartOptions = {
			'dataPoint1': {
				fillColor: '#F1EEF6',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint2': {
				fillColor: '#BDC9E1',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint3': {
				fillColor: '#74A9CF',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint4': {
				fillColor: '#0570B0',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			}
		};
		
		return new L.PieChartMarker(latlng, options);
	});
	
	addMarkers('Coxcomb Charts', -4.0, 14.0, 2.0, 5, function (latlng) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: 30,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			numberOfSides: 50,
			width: 10
		};

		options.data = {
			'dataPoint1': Math.random() * 20,
			'dataPoint2': Math.random() * 20,
			'dataPoint3': Math.random() * 20,
			'dataPoint4': Math.random() * 20
		};
		
		options.chartOptions = {
			'dataPoint1': {
				fillColor: '#EDF8FB',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint2': {
				fillColor: '#B2E2E2',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint3': {
				fillColor: '#66C2A4',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint4': {
				fillColor: '#238B45',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			}
		};

		return new L.CoxcombChartMarker(latlng, options);
	});
	
	addMarkers('Radial Bar Charts', -6.0, 14.0, 2.0, 5, function (latlng) {
		var colorValue = Math.random() * 360;
		var options = {
			color: '#000',
			weight: 1,
			fillColor: 'hsl(' + colorValue + ',100%,50%)',
			radius: 20,
			fillOpacity: 0.7,
			rotation: 0.0,
			position: {
				x: 0,
				y: 0
			},
			offset: 0,
			numberOfSides: 50,
			width: 10
		};

		options.data = {
			'dataPoint1': Math.random() * 20,
			'dataPoint2': Math.random() * 20,
			'dataPoint3': Math.random() * 20,
			'dataPoint4': Math.random() * 20
		};
		
		options.chartOptions = {
			'dataPoint1': {
				fillColor: '#FEE5D9',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint2': {
				fillColor: '#FCAE91',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint3': {
				fillColor: '#FB6A4A',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			},
			'dataPoint4': {
				fillColor: '#CB181D',
				minValue: 0,
				maxValue: 20,
				maxHeight: 20,
				displayText: function (value) {
					return value.toFixed(2);
				}
			}
		};

		return new L.RadialBarChartMarker(latlng, options);
	});
	
	addMarkers('Stacked Regular Polygons', -8.0, 14.0, 2.0, 5, function (latlng, index) {
		var stackedOptions = {
			chartOptions: {
				'dataPoint4': {
					fillColor: 'rgb(255, 255, 212)',
					minValue: 0,
					maxValue: 20,
					maxRadius: 20,
					displayText: function (value) {
						return value.toFixed(2);
					}
				},
				'dataPoint3': {
					fillColor: 'rgb(254, 217, 142)',
					minValue: 0,
					maxValue: 20,
					maxRadius: 20,
					displayText: function (value) {
						return value.toFixed(2);
					}
				},
				'dataPoint2': {
					fillColor: 'rgb(254, 153, 41)',
					minValue: 0,
					maxValue: 20,
					maxRadius: 20,
					displayText: function (value) {
						return value.toFixed(2);
					}
				},
				'dataPoint1': {
					fillColor: 'rgb(204, 76, 2)',
					minValue: 0,
					maxValue: 20,
					maxRadius: 20,
					displayText: function (value) {
						return value.toFixed(2);
					}
				}
			},
			data: {
				dataPoint1: 5,
				dataPoint2: 3,
				dataPoint3: 8,
				dataPoint4: 15
			},
			color: '#000000',
			weight: 1,
			numberOfSides: 6,
			rotation: 0
		};
		
		stackedOptions.data = {};
		
		for (var i = 1; i <= 4; ++i) {
			stackedOptions.data['dataPoint' + i] = Math.random() * 20;
		};
		
		return new L.StackedRegularPolygonMarker(latlng, stackedOptions);
	});

	var directions = Object.keys(L.CalloutLine.DIRECTION);
	
	console.log(directions);
	
	var styles = Object.keys(L.CalloutLine.LINESTYLE);
	var xOffsets = [5, -5, 5, -5];
	var yOffsets = [-5, -5, 5, 5];
	
	var calloutGenerator = function (style) {
		return function (latlng, index) {
			var direction = directions[index % directions.length];
		
			return new L.Callout(latlng, {
				direction: L.CalloutLine.DIRECTION[direction],
				lineStyle: L.CalloutLine.LINESTYLE[style],
				numberOfSides: 3,
				arrow: true,
				color: '#C0C0C0',
				fillColor: '#C0C0C0',
				position: new L.Point(xOffsets[index], yOffsets[index]),
				size: new L.Point(40, 40),
				icon: new L.DivIcon({
					iconSize: new L.Point(28, 14),
					html: direction.toUpperCase(),
					className: 'callout-text'
				})
			});
		}
	};
	
	addMarkers('Callouts (Arced)', -1.0, 28.0, 0.0, 4, calloutGenerator(styles[0]));
	addMarkers('Callouts (Angled)', -4.0, 28.0, 0.0, 4, calloutGenerator(styles[1]));
	addMarkers('Callouts (Straight)', -7.0, 28.0, 0.0, 4, calloutGenerator(styles[2]));

});